from massmigration.migrations import MapperMigration


class Migration(MapperMigration):
    """ YOUR DESCRIPTION HERE. This will appear in the Django admin. """

    # This can be set to make a migration run on a specific backend, rather than the one that's
    backend: str = None
    # specified in the Django settings

    # If you need to configure the backend differently for each migration, this is a place for
    # passing parameters to it.
    backend_params: dict = {}

    # This can be set to specify the list of database aliases on which the migration can be applied.
    # The migration is not forced on a specific DB but rather the `db_alias` for the DB is passed to `get_queryset`
    # allowing to customise what is retrieved by the migration.
    # If None the migration can be applied to all databases.
    allowed_db_aliases: list = None


    dependencies = [{% for dependency in dependencies %}
        ("{{dependency.0}}", "{{dependency.1}}"),{% endfor %}
    ]

    def get_queryset(self, db_alias):

        # PUT YOUR CODE HERE.
        # It must return a queryset for the objects which you wish to perform the operation on.

        raise NotImplementedError

    def operation(self, obj, db_alias):

        # PUT YOUR CODE HERE.
        # It should perform the operation on the given object which will be an instance from the
        # queryset.

        raise NotImplementedError



# Third party
from collections import OrderedDict
from django.contrib import messages
from django.http import Http404
from django.shortcuts import redirect, render

# Mass Migration
from massmigration.exceptions import DependentMigrationNotApplied
from massmigration.loader import store
from massmigration.migrations import get_all_db_aliases
from massmigration.models import MigrationRecord
from massmigration.utils.permissions import superuser_required


@superuser_required()
def manage_migrations(request):
    """ A page to manage mass migrations. """
    migrations = store.all
    available_db_aliases = get_all_db_aliases()

    # Load the migration in bulk from every db to avoid a separate query for each one
    migration_records_by_db_alias = {
        db_alias: MigrationRecord.objects.using(db_alias).in_bulk()
        for db_alias in available_db_aliases
    }

    for migration in migrations:
        migration.records_map = OrderedDict()

        for db_alias in available_db_aliases:
            migration.records_map[db_alias] = {
                "is_allowed_on_db_alias": db_alias in migration.get_allowed_db_aliases(),
                "record": migration_records_by_db_alias[db_alias].get(migration.key, None)
            }

    context = {
        "migrations": migrations,
        "available_db_aliases": available_db_aliases,
    }
    return render(request, "massmigration/manage_migrations.html", context)


@superuser_required()
def run_migration(request, key, db_alias):
    """ Trigger the running of a migration. """
    migration = store.by_key.get(key)
    if not migration:
        raise Http404(f"Migration with key '{key}' not found.")

    record = migration.get_migration_record(db_alias)

    if record:
        messages.error(request, f"Migration '{key}' for db <{db_alias}> has already been started.")
        return redirect("massmigration_manage")
    try:
        migration.check_dependencies(db_alias)
    except DependentMigrationNotApplied as error:
        messages.error(request, str(error))
        return redirect("massmigration_manage")

    if request.method == "POST":
        migration.launch(db_alias)
        messages.success(request, f"Migration '{key}' started.")
        return redirect("massmigration_manage")

    # else...
    context = {
        "migration": migration,
        "db_alias": db_alias,
    }
    return render(request, "massmigration/run_migration.html", context)


@superuser_required()
def migration_detail(request, key, db_alias):
    """ View the details of a single migration. """
    # TODO(https://github.com/adamalton/django-mass-migration/issues/4):
    # The migration detail view should collate the state of the Migration
    # across all allowed_db_aliases.
    migration = store.by_key.get(key)
    if not migration:
        raise Http404(f"Migration with key {key} not found.")

    record = MigrationRecord.objects.using(db_alias).filter(key=key).first()
    dependencies = []
    dependency_keys = [MigrationRecord.key_from_name_tuple(x) for x in migration.dependencies]
    dependency_records_by_key = MigrationRecord.objects.using(db_alias).in_bulk(dependency_keys)
    for dep_key in dependency_keys:
        dependencies.append({
            "key": dep_key,
            "record": dependency_records_by_key.get(dep_key),
        })
    context = {
        "migration": migration,
        "record": record,
        "dependencies": dependencies,
        "db_alias": db_alias,
    }
    return render(request, "massmigration/migration_detail.html", context)


@superuser_required()
def delete_migration(request, key, db_alias):
    """ Delete a migration which has already started or has errored. """
    migration = store.by_key.get(key)
    if not migration:
        raise Http404(f"Migration with key {key} not found.")

    record = migration.get_migration_record(db_alias)

    if not record:
        messages.error(
            request,
            f"Can't delete migration '{key}' for db <{db_alias}> because no record for it exists. "
            "Maybe you deleted it already?"
        )
        return redirect("massmigration_manage")

    if request.method == "POST":
        record.delete()
        messages.success(request, f"Deleted record for migration '{key}")
        return redirect("massmigration_manage")

    # else...
    context = {
        "migration": migration,
        "db_alias": db_alias,
    }
    return render(request, "massmigration/delete_migration.html", context)



"""
Utilities for allowing you to prevent particular code functions from being run unless specific
migrations have been applied.
"""

# Standard library
from functools import wraps
import logging

# Third party
from django.http import HttpResponse

# Djangae Migrations
from .exceptions import DbAliasNotAllowed, RequiredMigrationNotApplied
from .loader import store
from .models import MigrationRecord
from .utils.test import in_tests

logger = logging.getLogger(__name__)

APPLIED_MIGRATIONS_CACHE = {}


def get_migration_key(migration_id_str_or_tuple):
    """
    Get the migration key from the provided migration ID string or tuple.

    Args:
        migration_id_str_or_tuple (str or tuple): The migration ID string or tuple.

    Returns:
        str: The migration key.

    Raises:
        ValueError: If the provided migration key doesn't exists.
    """
    if isinstance(migration_id_str_or_tuple, (list, tuple)):
        key = MigrationRecord.key_from_name_tuple(migration_id_str_or_tuple)
    else:
        key = migration_id_str_or_tuple

    if key not in store.by_key:
        raise ValueError(
            f"Provided migration with key '{key}' not found."
            f"Available keys are {', '.join(store.by_key.keys())}"
        )

    return key


def migration_is_applied(migration_identifier, db_alias):
    """ Tells you whether or not the specified migration has been applied to the DB.
        Positive (True) responses are cached to avoid repeated DB queries.
    """
    migration_key = get_migration_key(migration_identifier)
    cache_key = (get_migration_key(migration_identifier), db_alias)
    try:
        return APPLIED_MIGRATIONS_CACHE[cache_key]
    except KeyError:
        applied = MigrationRecord.objects.using(db_alias).filter(key=migration_key, is_applied=True).exists()
        if applied:
            APPLIED_MIGRATIONS_CACHE[cache_key] = True
            return True
    return False


def requires_migration(migration_identifier, db_aliases=[], is_view=False, skip_in_tests=True):
    """ Function decorator which prevents the function being run if the specified migration is not
        applied.
    """

    def decorator(function):
        @wraps(function)
        def replacement(*args, **kwargs):
            def enforce_migration():
                key = get_migration_key(migration_identifier)

                migration = store.by_key[key]

                allowed_db_aliases = migration.get_allowed_db_aliases()

                # By default if db_alias is not specified, we assume the migration needs to be applied on all the allowed_db_aliases
                # specified in the migration
                if not db_aliases:
                    required_migrations_db_aliases = allowed_db_aliases
                else:
                    if any([db_alias not in allowed_db_aliases for db_alias in db_aliases]):
                        raise DbAliasNotAllowed(
                            f"requires_migration decorator improperly configured. "
                            f"It requires the migration <{migration_identifier}> to have run on <{', '.join(db_aliases)}> "
                            f"while the allowed databases are <{', '.join(allowed_db_aliases)}>."
                        )
                    else:
                        required_migrations_db_aliases = db_aliases

                if not (skip_in_tests and in_tests()):
                    if not all([migration_is_applied(migration_identifier, db_alias) for db_alias in required_migrations_db_aliases]):
                        raise RequiredMigrationNotApplied(
                            f"Migration '{function}'' requires migration {migration_identifier} which has "
                            "not been applied."
                        )

            if is_view:
                try:
                    enforce_migration()
                except RequiredMigrationNotApplied:
                    logger.error(
                        "View function '%s' requires migration '%s' which has not yet been applied.",
                        function, migration_identifier
                    )
                    return HttpResponse(
                        "This resource requires data changes which have not yet been made.",
                        status=503,
                    )
            else:
                enforce_migration()

            return function(*args, **kwargs)
        return replacement
    return decorator


def view_requires_migration(migration_identifier, db_aliases=[], skip_in_tests=True):
    """ Same as `requires_migration`, but for view functions. Returns a 503 status HttpResponse
        rather than raising an exception.
    """
    return requires_migration(migration_identifier, is_view=True, db_aliases=db_aliases, skip_in_tests=skip_in_tests)


# Third party
from django.conf import settings
from django.db import transaction as django_transaction


def get_transaction(db_alias):
    engine = settings.DATABASES[db_alias]["ENGINE"]
    transaction = django_transaction
    if engine == "gcloudc.db.backends.datastore":
        try:
            from gcloudc.db import transaction as gcloudc_transaction
            transaction = gcloudc_transaction
        except ImportError:
            # Newer versions of gcloudc use Django transactions.
            pass
    return transaction



#!/usr/bin/env python

""" Make and push a tag which will trigger the release process (which is a GitHub Action).
    Usage:
        ./release.py kind
    where `kind` is major/minor/bugfix.
"""

import subprocess
import sys

KINDS = ("major", "minor", "bugfix")


def get_kind():
    kind = sys.argv[-1]
    while kind.lower() not in KINDS:
        kind = input(f"What kind of release? [{'/'.join(KINDS)}] ")
    return kind


def get_latest_tag():
    tags = subprocess.check_output(["git", "tag", "--list", "v*"]).split()
    return tags[-1].decode("utf8")


def increment_tag(tag, kind):
    """ Increment the relevant part of the tag for the given kind of release. """
    major, minor, bugfix = tag.replace("v", "").split(".")
    if kind == "major":
        major = int(major) + 1
        minor = 0
        bugfix = 0
    elif kind == "minor":
        minor = int(minor) + 1
        bugfix = 0
    elif kind == "bugfix":
        bugfix = int(bugfix) + 1
    return f"v{major}.{minor}.{bugfix}"


def push_tag(tag_name):
    subprocess.check_output(["git", "tag", tag_name])
    subprocess.check_output(["git", "push", "origin", tag_name])


def main():
    tag_name = increment_tag(get_latest_tag(), get_kind())
    print(f"Pushing new tag: {tag_name}")
    push_tag(tag_name)
    print("Done")


if __name__ == "__main__":
    main()
